
//*******************************************************
// Unparser for VRML++ Preprocessor
// Copyright (C) 1996 Stephan Diehl
// 
//
// Converts Parse-Tree into VRML text
//
//*******************************************************

#include <iostream.h>
#include <stdio.h>
#include <malloc.h>
#include <assert.h>
#include <list.h>
#include <set.h>
#include <stack.h>
#include <string.h>

#include "vrmlnode.h"
#include "tokens.h"

extern char* whichscript;

/* The constants defined in parsetree.h (0 to ...) and
   tokens.h (258 to ...) have to be disjoined !!! */

void MakeDynRouteScript(list<ParseTreeNode*> ,int);
void RouteEventToNodeEventPair(ParseTreeNode*,char*);

void unparse(ParseTreeNode*, int);

extern int fieldType(const char *type);
extern char* nodeTypeToString(int type);

#define fieldTypeToString(t) nodeTypeToString(fieldType(t))

void unparse(ParseTreeNode* node)
 { if (node) unparse(node,0);
 }

void indent(int d) { while(d--) { cout << " "; }}

void unparse(ParseTreeNode* node, int d)
{ list<ParseTreeNode*> kids=node->getChildren();
  list<ParseTreeNode*>::iterator i=kids.begin();
  
  switch(node->type)
   { case PT_NONE:	break;
     case PT_ID:	cout << node->name << " ";
                        break;
     case PT_DECS: 	if (!kids.empty()) for(;i!=kids.end();i++)
                            { unparse(*i,d+2);
                              cout << "\n\n";
                              indent(d);
                            }
                        break;
     case PT_DEF:	cout << "DEF " << node->name << " ";
                        unparse(*i,d+2);
                        break;
     case PT_USE:	cout << "USE " << node->name << "\n";
                        indent(d);
                        break;
     case PT_PROTO:	cout << "PROTO " << node->name << " [ \n";
                        indent(d+2);
                        unparse(*i,d+2);
                        cout << "]\n";
                        indent(d+2);
                        cout << "{\n";
                        indent(d+2);
                        i++;
                        unparse(*i,d+2);
                        cout << "}\n";
                        indent(d);
                        break;
     case PT_EXTPROTO:  cout << "PROTO " << node->name << " [ \n";
                        indent(d+2);
                        unparse(*i,d+2);
                        cout << "]\n";
                        indent(d+2);
                        i++;
                        unparse(*i,d+2);
                        break;
     case PT_INTERFACE:	if (!kids.empty()) for(;i!=kids.end();i++)
                          { unparse(*i,d);
                          }    
                        break;
     case PT_EVENTIN:	cout << "  eventIn " << fieldTypeToString((*i)->name) << " ";
                        cout << node->name << "\n";
                        indent(d);
                        break;
     case PT_EVENTOUT:	cout << "  eventOut " << fieldTypeToString((*i)->name) << " ";
                        cout << node->name << "\n";
                        indent(d);
                        break;
     case PT_FIELD:	cout << "  field " << fieldTypeToString((*i)->name) << " ";
                        cout << node->name << " ";
                        if (i!=kids.end())
                           { i++;
                             unparse(*i,d+2);
                           }
                        cout << "\n";
			indent(d);
                        break;
     case PT_EXPFIELD:	cout << "  exposedField " << fieldTypeToString((*i)->name) << " ";
                        cout  << node->name << " ";
                        if (i!=kids.end())
                           { i++;
                             unparse(*i,d+2);
                           }
                        cout << "\n";
                        indent(d);
			break;
     case PT_EXTINTERFACE:	if (!kids.empty()) for(;i!=kids.end();i++)
                                 { unparse(*i,d);
                                 }
			break;
     case PT_ROUTE:	cout << "ROUTE "; 
                        unparse(*i,d);
                        i++;
                        cout << " TO ";
                        unparse(*i,d);
                        cout << "\n";
                        indent(d);
 			break;
     case PT_R_EVENT_A:	cout << node->name;
			break;
     case PT_ROUTE_EVENT_F_SF:
     case PT_ROUTE_EVENT_F_MF:
     case PT_ROUTE_EVENT_EX_SF:
     case PT_ROUTE_EVENT_EX_MF:
     case PT_ROUTE_EVENT_EI_SF:
     case PT_ROUTE_EVENT_EI_MF:
     case PT_ROUTE_EVENT_EO_SF:
     case PT_ROUTE_EVENT_EO_MF:
     case PT_R_EVENT_C:	
     case PT_R_EVENT_B:	cout << node->name << ".";
                        unparse(*i,d);
                        break;
     case PT_NODE:	cout << node->name << " {\n";
                        indent(d+2);
                        unparse(*i,d+2);
                        cout << "}\n";
                        indent(d);
			break;
     case PT_GUTS:	if (!kids.empty()) for(;i!=kids.end();i++)
                          { unparse(*i,d);
                            cout << "\n";
                            indent(d);
                          }
			break;
     case PT_IDENTIFIER:	cout << "  " << node->name << " ";
                                unparse(*i,d);
				break;
     case PT_EVENTIN_IS:	cout << "  eventIn " << fieldTypeToString((*i)->name) << " ";
                                cout << node->name << " IS ";
                                i++;
                                unparse(*i,d);
                                cout << "\n";
                                indent(d);
                                break;
     case PT_EVENTOUT_IS:	cout << "  eventOut " << fieldTypeToString((*i)->name) << " ";
                                cout << node->name << " IS ";
                                i++;
                                unparse(*i,d);
                                cout << "\n";
                                indent(d);
      				break;
     case PT_CLASS:	cout << "CLASS " << node->name << "[\n";
                        indent(d+2);
                        unparse(*i,d+2);
                        cout << "] EXTENDS ";
                        i++;
                        unparse(*i,d+2);
                        cout << "{\n";
                        indent(d+2);
                        i++;
                        unparse(*i,d+2);
                        cout << "}\n";
                        indent(d);
			break;
     case PT_EXTCLASS:	cout << "CLASS " << node->name << "[\n";
                        indent(d+2);
                        unparse(*i,d+2);
                        cout << "] EXTENDS ";
                        i++;
                        unparse(*i,d+2);
                        i++;
                        cout << "\n";
                        indent(d+2);
                        unparse(*i,d+2);
			break;
     case PT_SUPERS:	if (!kids.empty()) for(;i!=kids.end();i++)
                         { unparse(*i,d);
                         }
			break;
     case PT_VALUE:	/* not used */
			break;
     case PT_SFNODE:	unparse(*i,d);
                        break;
     case PT_SFN_NULL:	cout << "NULL ";
                        break;
     case PT_MFNODE:	unparse(*i,d);
			break;
     case PT_MFNODE2:	cout << "[ ";
                        unparse(*i,d+2);
                        cout << "] ";
			break;
     case PT_IS:	cout << "IS ";
                        unparse(*i,d+2);
			break;
     case SFBOOL:
     case SFCOLOR:
     case MFCOLOR:
     case SFFLOAT:
     case MFFLOAT:
     case SFIMAGE:
     case SFINT32:
     case MFINT32:
     case SFROTATION:
     case MFROTATION:
     case SFSTRING:
     case MFSTRING:
     case SFTIME:
     case SFVEC2F:
     case MFVEC2F:
     case SFVEC3F:
     case MFVEC3F:	cout << node->name << " ";
			break;
     case PT_DYN_ROUTE: // children are a list of dynamic routes
                        // and have to be translated into a script
                        MakeDynRouteScript(node->getChildren(),d);
                        break;
     default:	cerr << "Error when unparsing !!!\n";
            	cout << "\nError: unknown node in parse tree\n";
            	break;
   }
}

int findNameInList(char* name,list<ParseTreeNode*> & plist,
                   list<ParseTreeNode*>::iterator & i)
 { list<ParseTreeNode*>::iterator j=plist.begin();
   while ((*j)!=(*i))
    { if (strcmp((*j)->name,name)==0) return 1;
      j++;
    }
   return(0);
 }

void MakeDynRouteScript(list<ParseTreeNode*> dynNode, int d)
 { list<ParseTreeNode*>::iterator j=dynNode.begin();
   list<ParseTreeNode*> routes, names, events; 
   list<ParseTreeNode*>::iterator ni,ri,ei;
   
   routes=(*j)->getChildren();   
   j++;
   names=(*j)->getChildren();
   j++;
   events=(*j)->getChildren();

   cout << "\n";
   indent(d);
   cout << "DEF XXX_RR Script { directOutput TRUE\n";
   indent(d+2);

   for(ni=names.begin();ni!=names.end();ni++)
    { if (!findNameInList((*ni)->name,names,ni)) 
       { cout << "field SFNode xxx_node_"; 
         cout << (*ni)->name 
              << " USE " << (*ni)->name << "\n"; 
         indent(d+2);
       }
    }

// Do we have to connect script events to proto events by IS ???
   for(ei=events.begin();ei!=events.end();ei++)
    { if (!findNameInList((*ei)->name,events,ei)) 
        {
         switch ((*ei)->type)
          { case PT_ROUTE_EVENT_F_SF:  cout << "field SFNode "; 
                                       break;
            case PT_ROUTE_EVENT_F_MF:  cout << "field MFNode "; 
                                       break;
            case PT_ROUTE_EVENT_EX_SF: cout << "exposedField SFNode "; 
                                       break;
            case PT_ROUTE_EVENT_EX_MF: cout << "exposedField MFNode "; 
                                       break;
            case PT_ROUTE_EVENT_EI_SF: cout << "eventIn SFNode "; 
                                       break;
            case PT_ROUTE_EVENT_EI_MF: cout << "eventIn MFNode "; 
                                       break;
            case PT_ROUTE_EVENT_EO_SF: cout << "eventOut SFNode "; 
                                       break;
            case PT_ROUTE_EVENT_EO_MF: cout << "eventOut MFNode "; 
                                       break;
            default: cout << "Error, no field or event of type SFNode or MFNode : ";
          }
          cout << (*ei)->name
               << " IS " << (*ei)->name << "\n"; 
          indent(d+2);
        }
    }
  
   cout << "url \"" << whichscript << ": function initialize(value) { \n";
   
   indent(d+4);
   // now we add the new routes !!!
   for(ri=routes.begin();ri!=routes.end();ri++)
    { list<ParseTreeNode*> kids=(*ri)->getChildren();
      list<ParseTreeNode*>::iterator rr=kids.begin();
      ParseTreeNode *from,*to;
      char *mf_from="";
      char *mf_to="";

      from=(*rr); rr++; to=(*rr);

  /* In older version's of the WorldView browser
     arrays have been indexed starting at 1,
     in WorldView 2.0 indexing starts at 0 */ 
      switch (from->type)
       { case PT_ROUTE_EVENT_F_MF:
         case PT_ROUTE_EVENT_EX_MF:
         case PT_ROUTE_EVENT_EI_MF:
         case PT_ROUTE_EVENT_EO_MF: mf_from="[i]";
                                    cout << "var i; for(i=0;i<" 
                                         << from->name 
                                         << ".length;i++) {\n"; 
         default: ;
       }
      switch (to->type)
       { case PT_ROUTE_EVENT_F_MF:
         case PT_ROUTE_EVENT_EX_MF:
         case PT_ROUTE_EVENT_EI_MF:
         case PT_ROUTE_EVENT_EO_MF: mf_to="[j]";
                                    cout << "var j; for(j=0;j<" 
                                         << to->name 
                                         << ".length;j++) {\n"; 
         default: ;
       }
      if ((*ri)->type==PT_ROUTE) cout << "Browser.addRoute(";
      if ((*ri)->type==PT_UNROUTE) cout << "Browser.deleteRoute(";
      RouteEventToNodeEventPair(from,mf_from);
      cout << ",";
      RouteEventToNodeEventPair(to,mf_to);
      cout << ");\n";
      indent(d+4);
      switch (from->type)
       { case PT_ROUTE_EVENT_F_MF:
         case PT_ROUTE_EVENT_EX_MF:
         case PT_ROUTE_EVENT_EI_MF:
         case PT_ROUTE_EVENT_EO_MF: cout << "}\n";
                                    indent(d+4);
       }
      switch (to->type)
       { case PT_ROUTE_EVENT_F_MF:
         case PT_ROUTE_EVENT_EX_MF:
         case PT_ROUTE_EVENT_EI_MF:
         case PT_ROUTE_EVENT_EO_MF: cout << "}\n";
                                    indent(d+4);
       }
    }
   indent(d+2);
   cout << "}\"\n";
   indent(d);
   cout << "}\n";
 }
 
void RouteEventToNodeEventPair(ParseTreeNode* pt,char* prefix)
 { switch (pt->type)
     { case PT_R_EVENT_A:  cout << "XXX_RR,'" 
                                << pt->name 
                                << "'";
                           break;
       case PT_ROUTE_EVENT_F_MF:
       case PT_ROUTE_EVENT_EX_MF:
       case PT_ROUTE_EVENT_EI_MF:
       case PT_ROUTE_EVENT_EO_MF: cout << pt->name 
                                       << prefix 
                                       << ",'" 
                                       << (*(pt->getChildren().begin()))->name
                                       << "'";
                  break;
       case PT_ROUTE_EVENT_F_SF:
       case PT_ROUTE_EVENT_EX_SF:
       case PT_ROUTE_EVENT_EI_SF:
       case PT_ROUTE_EVENT_EO_SF:
       case PT_R_EVENT_B:  cout << pt->name 
                                << ",'" 
                                << (*(pt->getChildren().begin()))->name 
                                << "'";
                           break;
       case PT_R_EVENT_C:  cout << "xxx_node_" 
                                << pt->name 
                                << ",'" 
                                << (*(pt->getChildren().begin()))->name 
                                << "'";
                           break;
       default: cout << "Error:" << pt->type;
     }
   return;
 }
